@@ -21,6 +21,77 @@ defmodule CodeTest do
 
   Code.eval_quoted(contents, [], file: "sample.ex", line: 13)
 
+  describe "with_diagnostics/2" do
+    test "captures warnings" do
+      assert {:warn, [%{message: "hello"}]} =
+               Code.with_diagnostics(fn ->
+                 IO.warn("hello")
+                 :warn
+               end)
+    end
+
+    test "captures and logs warnings" do
+      assert ExUnit.CaptureIO.capture_io(:stderr, fn ->
+               assert {:warn, [%{message: "hello"}]} =
+                        Code.with_diagnostics([log: true], fn ->
+                          IO.warn("hello")
+                          :warn
+                        end)
+             end) =~ "hello"
+    end
+
+    test "can be nested" do
+      assert {:warn, [%{message: "hello"}]} =
+               Code.with_diagnostics(fn ->
+                 IO.warn("hello")
+
+                 assert {:nested, [%{message: "world"}]} =
+                          Code.with_diagnostics(fn ->
+                            IO.warn("world")
+                            :nested
+                          end)
+
+                 :warn
+               end)
+    end
+
+    test "includes column information on unused variables" do
+      assert {_, [%{position: {1, 12}}]} =
+               Code.with_diagnostics(fn ->
+                 quoted = Code.string_to_quoted!("if true do var = :foo end", columns: true)
+                 Code.eval_quoted(quoted, [])
+               end)
+    end
+
+    test "includes column information on unused aliases" do
+      sample = """
+      defmodule CodeTest.UnusedAlias do
+        alias String.Chars
+      end
+      """
+
+      assert {_, [%{position: {2, 3}}]} =
+               Code.with_diagnostics(fn ->
+                 quoted = Code.string_to_quoted!(sample, columns: true)
+                 Code.eval_quoted(quoted, [])
+               end)
+    end
+
+    test "includes column information on unused imports" do
+      sample = """
+      defmodule CodeTest.UnusedImport do
+        import URI
+      end
+      """
+
+      assert {_, [%{position: {2, 3}}]} =
+               Code.with_diagnostics(fn ->
+                 quoted = Code.string_to_quoted!(sample, columns: true)
+                 Code.eval_quoted(quoted, [])
+               end)
+    end
+  end
+
   describe "eval_string/1,2,3" do
     test "correctly evaluates a string of code" do
       assert Code.eval_string("1 + 2") == {3, []}
@@ -40,10 +111,6 @@ defmodule CodeTest do
                Code.eval_string("a = (try do (raise \"hello\") rescue e -> e end)")
     end
 
-    test "supports the :requires option" do
-      assert Code.eval_string("Kernel.if true, do: :ok", [], requires: [Z, Kernel]) == {:ok, []}
-    end
-
     test "returns bindings from a different context" do
       assert Code.eval_string("var!(a, Sample) = 1") == {1, [{{:a, Sample}, 1}]}
     end
@@ -67,18 +134,6 @@ defmodule CodeTest do
                {"1", [{:c, 2}, {:b, "1"}, {:a, 1}]}
     end
 
-    test "with many options" do
-      options = [
-        functions: [{Kernel, [is_atom: 1]}],
-        macros: [{Kernel, [and: 2]}],
-        aliases: [{K, Kernel}],
-        requires: [Kernel]
-      ]
-
-      code = "is_atom(:foo) and K.is_list([])"
-      assert Code.eval_string(code, [], options) == {true, []}
-    end
-
     test "keeps caller in stacktrace" do
       try do
         Code.eval_string("<<a::size(b)>>", [a: :a, b: :b], file: "myfile")
@@ -127,28 +182,66 @@ defmodule CodeTest do
         end)
 
       assert output =~ "incompatible types"
+    after
+      :code.purge(CodeTest.CheckerWarning)
+      :code.delete(CodeTest.CheckerWarning)
+    end
+
+    test "captures checker diagnostics" do
+      {{{:module, _, _, _}, _}, diagnostics} =
+        Code.with_diagnostics(fn ->
+          Code.eval_string(File.read!(fixture_path("checker_warning.exs")), [])
+        end)
+
+      assert [%{message: "incompatible types:" <> _}] = diagnostics
+    after
+      :code.purge(CodeTest.CheckerWarning)
+      :code.delete(CodeTest.CheckerWarning)
+    end
+
+    test "formats diagnostic file paths as relatives" do
+      {_, diagnostics} =
+        Code.with_diagnostics(fn ->
+          try do
+            Code.eval_string("x", [])
+          rescue
+            e -> e
+          end
+        end)
+
+      assert [
+               %{
+                 message: "undefined variable \"x\"",
+                 position: 1,
+                 file: "nofile",
+                 stacktrace: [],
+                 severity: :error
+               }
+             ] = diagnostics
     end
   end
 
-  test "eval_quoted/1" do
-    assert Code.eval_quoted(quote(do: 1 + 2)) == {3, []}
-    assert CodeTest.Sample.eval_quoted_info() == {CodeTest.Sample, "sample.ex", 13}
-  end
+  describe "eval_quoted/1" do
+    test "evaluates expression" do
+      assert Code.eval_quoted(quote(do: 1 + 2)) == {3, []}
+      assert CodeTest.Sample.eval_quoted_info() == {CodeTest.Sample, "sample.ex", 13}
+    end
 
-  test "eval_quoted/2 with %Macro.Env{} at runtime" do
-    alias :lists, as: MyList
-    quoted = quote(do: MyList.flatten([[1, 2, 3]]))
+    test "with %Macro.Env{} at runtime" do
+      alias :lists, as: MyList
+      quoted = quote(do: MyList.flatten([[1, 2, 3]]))
 
-    assert Code.eval_quoted(quoted, [], __ENV__) == {[1, 2, 3], []}
+      assert Code.eval_quoted(quoted, [], __ENV__) == {[1, 2, 3], []}
 
-    # Let's check it discards tracers since the lexical tracker is explicitly nil
-    assert Code.eval_quoted(quoted, [], %{__ENV__ | tracers: [:bad]}) == {[1, 2, 3], []}
-  end
+      # Let's check it discards tracers since the lexical tracker is explicitly nil
+      assert Code.eval_quoted(quoted, [], %{__ENV__ | tracers: [:bad]}) == {[1, 2, 3], []}
+    end
 
-  test "eval_quoted/2 with %Macro.Env{} at compile time" do
-    defmodule CompileTimeEnv do
-      alias String.Chars
-      {"foo", []} = Code.eval_string("Chars.to_string(:foo)", [], __ENV__)
+    test "with %Macro.Env{} at compile time" do
+      defmodule CompileTimeEnv do
+        alias String.Chars
+        {"foo", []} = Code.eval_string("Chars.to_string(:foo)", [], __ENV__)
+      end
     end
   end
 
@@ -160,101 +253,137 @@ defmodule CodeTest do
     end
   end
 
-  test "eval_quoted_with_env/3" do
-    alias :lists, as: MyList
-    quoted = quote(do: MyList.flatten([[1, 2, 3]]))
-    env = Code.env_for_eval(__ENV__)
-    assert Code.eval_quoted_with_env(quoted, [], env) == {[1, 2, 3], [], env}
+  describe "eval_quoted_with_env/3" do
+    test "returns results, bindings, and env" do
+      alias :lists, as: MyList
+      quoted = quote(do: MyList.flatten([[1, 2, 3]]))
+      env = Code.env_for_eval(__ENV__)
+      assert Code.eval_quoted_with_env(quoted, [], env) == {[1, 2, 3], [], env}
 
-    quoted = quote(do: alias(:dict, as: MyDict))
-    {:dict, [], env} = Code.eval_quoted_with_env(quoted, [], env)
-    assert Macro.Env.fetch_alias(env, :MyDict) == {:ok, :dict}
-  end
-
-  test "eval_quoted_with_env/3 with vars" do
-    env = Code.env_for_eval(__ENV__)
-    {1, [x: 1], env} = Code.eval_quoted_with_env(quote(do: var!(x) = 1), [], env)
-    assert Macro.Env.vars(env) == [{:x, nil}]
-  end
-
-  test "eval_quoted_with_env/3 with pruning" do
-    env = Code.env_for_eval(__ENV__)
-
-    fun = fn quoted, binding ->
-      {_, binding, env} = Code.eval_quoted_with_env(quoted, binding, env, prune_binding: true)
-      {binding, Macro.Env.vars(env)}
+      quoted = quote(do: alias(:dict, as: MyDict))
+      {:dict, [], env} = Code.eval_quoted_with_env(quoted, [], env)
+      assert Macro.Env.fetch_alias(env, :MyDict) == {:ok, :dict}
     end
 
-    assert fun.(quote(do: 123), []) == {[], []}
-    assert fun.(quote(do: 123), x: 2, y: 3) == {[], []}
-
-    assert fun.(quote(do: var!(x) = 1), []) == {[x: 1], [x: nil]}
-    assert fun.(quote(do: var!(x) = 1), x: 2, y: 3) == {[x: 1], [x: nil]}
-
-    assert fun.(quote(do: var!(x, :foo) = 1), []) == {[{{:x, :foo}, 1}], [x: :foo]}
-    assert fun.(quote(do: var!(x, :foo) = 1), x: 2, y: 3) == {[{{:x, :foo}, 1}], [x: :foo]}
-
-    assert fun.(quote(do: var!(x, :foo) = 1), [{{:x, :foo}, 2}, {{:y, :foo}, 3}]) ==
-             {[{{:x, :foo}, 1}], [x: :foo]}
-
-    assert fun.(quote(do: fn -> var!(x, :foo) = 1 end), []) == {[], []}
-    assert fun.(quote(do: fn -> var!(x, :foo) = 1 end), x: 1, y: 2) == {[], []}
-
-    assert fun.(quote(do: fn -> var!(x) end), x: 2, y: 3) == {[x: 2], [x: nil]}
-
-    assert fun.(quote(do: fn -> var!(x, :foo) end), [{{:x, :foo}, 2}, {{:y, :foo}, 3}]) ==
-             {[{{:x, :foo}, 2}], [x: :foo]}
-  end
-
-  defmodule Tracer do
-    def trace(event, env) do
-      send(self(), {:trace, event, env})
-      :ok
+    test "manages env vars" do
+      env = Code.env_for_eval(__ENV__)
+      {1, [x: 1], env} = Code.eval_quoted_with_env(quote(do: var!(x) = 1), [], env)
+      assert Macro.Env.vars(env) == [{:x, nil}]
     end
-  end
 
-  test "eval_quoted_with_env/3 with tracing and pruning" do
-    env = %{Code.env_for_eval(__ENV__) | tracers: [Tracer], function: nil}
-    binding = [x: 1, y: 2, z: 3]
+    test "prunes vars" do
+      env = Code.env_for_eval(__ENV__)
 
-    quoted =
-      quote do
-        defmodule Elixir.CodeTest.TracingPruning do
-          var!(y) = :updated
-          var!(y)
-          var!(x)
-        end
+      fun = fn quoted, binding ->
+        {_, binding, env} = Code.eval_quoted_with_env(quoted, binding, env, prune_binding: true)
+        {binding, Macro.Env.vars(env)}
       end
 
-    {_, binding, env} = Code.eval_quoted_with_env(quoted, binding, env, prune_binding: true)
-    assert Enum.sort(binding) == []
-    assert env.versioned_vars == %{}
+      assert fun.(quote(do: 123), []) == {[], []}
+      assert fun.(quote(do: 123), x: 2, y: 3) == {[], []}
 
-    assert_receive {:trace, {:on_module, _, _}, %{module: CodeTest.TracingPruning} = trace_env}
-    assert trace_env.versioned_vars == %{{:result, Kernel} => 5, {:x, nil} => 1, {:y, nil} => 4}
+      assert fun.(quote(do: var!(x) = 1), []) == {[x: 1], [x: nil]}
+      assert fun.(quote(do: var!(x) = 1), x: 2, y: 3) == {[x: 1], [x: nil]}
+
+      assert fun.(quote(do: var!(x, :foo) = 1), []) == {[{{:x, :foo}, 1}], [x: :foo]}
+      assert fun.(quote(do: var!(x, :foo) = 1), x: 2, y: 3) == {[{{:x, :foo}, 1}], [x: :foo]}
+
+      assert fun.(quote(do: var!(x, :foo) = 1), [{{:x, :foo}, 2}, {{:y, :foo}, 3}]) ==
+               {[{{:x, :foo}, 1}], [x: :foo]}
+
+      assert fun.(quote(do: fn -> var!(x, :foo) = 1 end), []) == {[], []}
+      assert fun.(quote(do: fn -> var!(x, :foo) = 1 end), x: 1, y: 2) == {[], []}
+
+      assert fun.(quote(do: fn -> var!(x) end), x: 2, y: 3) == {[x: 2], [x: nil]}
+
+      assert fun.(quote(do: fn -> var!(x, :foo) end), [{{:x, :foo}, 2}, {{:y, :foo}, 3}]) ==
+               {[{{:x, :foo}, 2}], [x: :foo]}
+    end
+
+    defmodule Tracer do
+      def trace(event, env) do
+        send(self(), {:trace, event, env})
+        :ok
+      end
+    end
+
+    test "with tracing and pruning" do
+      env = %{Code.env_for_eval(__ENV__) | tracers: [Tracer], function: nil}
+      binding = [x: 1, y: 2, z: 3]
+
+      quoted =
+        quote do
+          defmodule Elixir.CodeTest.TracingPruning do
+            var!(y) = :updated
+            var!(y)
+            var!(x)
+          end
+        end
+
+      {_, binding, env} = Code.eval_quoted_with_env(quoted, binding, env, prune_binding: true)
+      assert Enum.sort(binding) == []
+      assert env.versioned_vars == %{}
+
+      assert_receive {:trace, {:on_module, _, _}, %{module: CodeTest.TracingPruning} = trace_env}
+      assert trace_env.versioned_vars == %{{:result, Kernel} => 5, {:x, nil} => 1, {:y, nil} => 4}
+    end
+
+    test "with defguard" do
+      require Integer
+      env = Code.env_for_eval(__ENV__)
+      quoted = quote do: Integer.is_even(1)
+      {false, binding, env} = Code.eval_quoted_with_env(quoted, [], env, prune_binding: true)
+      assert binding == []
+      assert Macro.Env.vars(env) == []
+    end
   end
 
-  test "eval_quoted_with_env/3 with defguard" do
-    require Integer
-    env = Code.env_for_eval(__ENV__)
-    quoted = quote do: Integer.is_even(1)
-    {false, binding, env} = Code.eval_quoted_with_env(quoted, [], env, prune_binding: true)
-    assert binding == []
-    assert Macro.Env.vars(env) == []
-  end
+  describe "compile_file/1" do
+    test "compiles the given path" do
+      assert Code.compile_file(fixture_path("code_sample.exs")) == []
+      refute fixture_path("code_sample.exs") in Code.required_files()
+    end
 
-  test "compile_file/1" do
-    assert Code.compile_file(fixture_path("code_sample.exs")) == []
-    refute fixture_path("code_sample.exs") in Code.required_files()
-  end
+    test "emits checker warnings" do
+      output =
+        ExUnit.CaptureIO.capture_io(:stderr, fn ->
+          Code.compile_file(fixture_path("checker_warning.exs"))
+        end)
 
-  test "compile_file/1 also emits checker warnings" do
-    output =
-      ExUnit.CaptureIO.capture_io(:stderr, fn ->
-        Code.compile_file(fixture_path("checker_warning.exs"))
-      end)
+      assert output =~ "incompatible types"
+    after
+      :code.purge(CodeTest.CheckerWarning)
+      :code.delete(CodeTest.CheckerWarning)
+    end
 
-    assert output =~ "incompatible types"
+    test "captures checker diagnostics" do
+      {[{CodeTest.CheckerWarning, _}], diagnostics} =
+        Code.with_diagnostics(fn ->
+          Code.compile_file(fixture_path("checker_warning.exs"))
+        end)
+
+      assert [%{message: "incompatible types:" <> _}] = diagnostics
+    after
+      :code.purge(CodeTest.CheckerWarning)
+      :code.delete(CodeTest.CheckerWarning)
+    end
+
+    test "captures checker diagnostics with logging" do
+      output =
+        ExUnit.CaptureIO.capture_io(:stderr, fn ->
+          {[{CodeTest.CheckerWarning, _}], diagnostics} =
+            Code.with_diagnostics([log: true], fn ->
+              Code.compile_file(fixture_path("checker_warning.exs"))
+            end)
+
+          assert [%{message: "incompatible types:" <> _}] = diagnostics
+        end)
+
+      assert output =~ "incompatible types"
+    after
+      :code.purge(CodeTest.CheckerWarning)
+      :code.delete(CodeTest.CheckerWarning)
+    end
   end
 
   test "require_file/1" do
@@ -333,6 +462,9 @@ defmodule CodeTest do
         end)
 
       assert output =~ "incompatible types"
+    after
+      :code.purge(CodeTest.CheckerWarning)
+      :code.delete(CodeTest.CheckerWarning)
     end
 
     test "works across lexical scopes" do
@@ -361,6 +493,30 @@ defmodule CodeTest do
     end
   end
 
+  test "ensure_all_loaded/1" do
+    assert Code.ensure_all_loaded([__MODULE__]) == :ok
+    assert Code.ensure_all_loaded([__MODULE__, Kernel]) == :ok
+
+    assert {:error, [error]} = Code.ensure_all_loaded([__MODULE__, Code.NoFile, __MODULE__])
+    assert error == {Code.NoFile, :nofile}
+  end
+
+  test "ensure_all_loaded!/1" do
+    assert Code.ensure_all_loaded!([__MODULE__]) == :ok
+    assert Code.ensure_all_loaded!([__MODULE__, Kernel]) == :ok
+
+    message = """
+    could not load the following modules:
+
+      * Code.NoFile due to reason :nofile
+      * Code.OtherNoFile due to reason :nofile\
+    """
+
+    assert_raise ArgumentError, message, fn ->
+      Code.ensure_all_loaded!([__MODULE__, Code.NoFile, Code.OtherNoFile])
+    end
+  end
+
   test "ensure_compiled/1" do
     assert Code.ensure_compiled(__MODULE__) == {:module, __MODULE__}
     assert Code.ensure_compiled(Code.NoFile) == {:error, :nofile}
@@ -405,15 +561,80 @@ defmodule Code.SyncTest do
 
   import PathHelpers
 
-  test "path manipulation" do
+  if :erlang.system_info(:otp_release) >= ~c"26" do
+    defp assert_cached(path) do
+      assert find_path(path) != :nocache
+    end
+
+    defp refute_cached(path) do
+      assert find_path(path) == :nocache
+    end
+
+    defp find_path(path) do
+      {:status, _, {:module, :code_server}, [_, :running, _, _, state]} =
+        :sys.get_status(:code_server)
+
+      [:state, _, _otp_root, paths | _] = Tuple.to_list(state)
+      {_, value} = List.keyfind(paths, to_charlist(path), 0)
+      value
+    end
+  else
+    defp assert_cached(_path), do: :ok
+    defp refute_cached(_path), do: :ok
+  end
+
+  test "prepend_path" do
     path = Path.join(__DIR__, "fixtures")
-    Code.prepend_path(path)
+    true = Code.prepend_path(path)
     assert to_charlist(path) in :code.get_path()
+    refute_cached(path)
+
+    true = Code.prepend_path(path, cache: true)
+    assert_cached(path)
 
     Code.delete_path(path)
     refute to_charlist(path) in :code.get_path()
   end
 
+  test "append_path" do
+    path = Path.join(__DIR__, "fixtures")
+    true = Code.append_path(path)
+    assert to_charlist(path) in :code.get_path()
+    refute_cached(path)
+
+    true = Code.append_path(path, cache: true)
+    assert_cached(path)
+
+    Code.delete_path(path)
+    refute to_charlist(path) in :code.get_path()
+  end
+
+  test "prepend_paths" do
+    path = Path.join(__DIR__, "fixtures")
+    :ok = Code.prepend_paths([path])
+    assert to_charlist(path) in :code.get_path()
+    refute_cached(path)
+
+    :ok = Code.prepend_paths([path], cache: true)
+    assert_cached(path)
+
+    Code.delete_paths([path])
+    refute to_charlist(path) in :code.get_path()
+  end
+
+  test "append_paths" do
+    path = Path.join(__DIR__, "fixtures")
+    :ok = Code.append_paths([path])
+    assert to_charlist(path) in :code.get_path()
+    refute_cached(path)
+
+    :ok = Code.append_paths([path], cache: true)
+    assert_cached(path)
+
+    Code.delete_paths([path])
+    refute to_charlist(path) in :code.get_path()
+  end
+
   test "purges compiler modules" do
     quoted = quote(do: Agent.start_link(fn -> :ok end))
     Code.compile_quoted(quoted)
