@@ -13,7 +13,7 @@ defmodule IEx.Broker do
   Finds the IEx server.
   """
   @spec shell :: shell()
-  # TODO: Use shell:whereis_shell() from Erlang/OTP 26+.
+  # TODO: Use shell:whereis() from Erlang/OTP 26+.
   def shell() do
     if user = Process.whereis(:user) do
       if user_drv = get_from_dict(user, :user_drv) do
@@ -65,27 +65,60 @@ defmodule IEx.Broker do
 
   The broker's PID is needed to support remote shells.
   """
-  @spec respond(pid, take_ref, boolean()) :: :ok | {:error, :refused | :already_accepted}
-  def respond(broker_pid, take_ref, true) do
-    GenServer.call(broker_pid, {:accept, take_ref, Process.group_leader()})
+  @spec respond(pid, take_ref, integer(), boolean()) ::
+          :ok | {:error, :refused | :already_accepted}
+  def respond(broker_pid, take_ref, counter, true) do
+    GenServer.call(broker_pid, {:accept, take_ref, Process.group_leader(), counter})
   end
 
-  def respond(broker_pid, take_ref, false) do
+  def respond(broker_pid, take_ref, _counter, false) do
     GenServer.call(broker_pid, {:refuse, take_ref})
   end
 
+  @doc """
+  Asks to IO if we want to take over.
+  """
+  def take_over?(location, whereami, opts) do
+    evaluator = opts[:evaluator] || self()
+    message = "Request to pry #{inspect(evaluator)} at #{location}#{whereami}"
+    interrupt = IEx.color(:eval_interrupt, "#{message}\nAllow? [Yn] ")
+    yes?(IO.gets(:stdio, interrupt))
+  end
+
+  defp yes?(string) do
+    is_binary(string) and String.trim(string) in ["", "y", "Y", "yes", "YES", "Yes"]
+  end
+
   @doc """
   Client requests a takeover.
   """
   @spec take_over(binary, iodata, keyword) ::
-          {:ok, server :: pid, group_leader :: pid} | {:error, :no_iex | :refused}
+          {:ok, server :: pid, group_leader :: pid, counter :: integer}
+          | {:error, :no_iex | :refused}
   def take_over(location, whereami, opts) do
-    case GenServer.whereis(@name) do
-      nil ->
-        {:error, :no_iex}
+    case take_over_existing(location, whereami, opts) do
+      {:error, :no_iex} ->
+        cond do
+          # TODO: Remove this check on Erlang/OTP 26+ and {:error, :no_iex}
+          not Code.ensure_loaded?(:prim_tty) ->
+            {:error, :no_iex}
 
-      _pid ->
-        GenServer.call(@name, {:take_over, location, whereami, opts}, :infinity)
+          take_over?(location, whereami, opts) ->
+            {:ok, IEx.cli(opts), Process.group_leader(), 1}
+
+          true ->
+            {:error, :refused}
+        end
+
+      other ->
+        other
+    end
+  end
+
+  defp take_over_existing(location, whereami, opts) do
+    case GenServer.whereis(@name) do
+      nil -> {:error, :no_iex}
+      _pid -> GenServer.call(@name, {:take_over, location, whereami, opts}, :infinity)
     end
   end
 
@@ -125,13 +158,13 @@ defmodule IEx.Broker do
     {:reply, :ok, state}
   end
 
-  def handle_call({:accept, {ref, _server_ref}, group_leader}, {server, _}, state) do
+  def handle_call({:accept, {ref, _server_ref}, group_leader, counter}, {server, _}, state) do
     case pop_in(state.takeovers[ref]) do
       {nil, state} ->
         {:reply, {:error, :already_accepted}, state}
 
       {{from, _}, state} ->
-        GenServer.reply(from, {:ok, server, group_leader})
+        GenServer.reply(from, {:ok, server, group_leader, counter})
         {:reply, :ok, state}
     end
   end
@@ -178,8 +211,11 @@ defmodule IEx.Broker do
 
   defp local_or_remote_shell() do
     Enum.find_value([node() | Node.list()], fn node ->
-      server = :rpc.call(node, IEx.Broker, :shell, [])
-      if is_pid(server), do: server
+      try do
+        :erpc.call(node, IEx.Broker, :shell, [])
+      catch
+        _, _ -> nil
+      end
     end)
   end
 end
